/*
############################################################################
# PFTH - This program implements a Forth interpreter.
#
# Copyright (c) 2012 Dave Hein
# MIT Licensed
############################################################################
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include "pfth.h"
#include "words.h"

#undef  DEBUG

#ifdef __PROPELLER__
#include "propwords.h"

#ifdef THIS_IS_MAIN
extern _Driver _SimpleSerialDriver;
extern _Driver _FileDriver;

_Driver *_driverlist[] = {
  &_SimpleSerialDriver,
  &_FileDriver,
  NULL
};
#endif
#else
#include <termios.h>
#define THIS_IS_MAIN
#endif

// Push a value onto the return stack
void FAST pushret(int value)
{
    retstk[retidx++] = value;
}

// Pop a value from the return stack
int FAST popret(void)
{
    if (retidx == 0) return RETURN_ERROR;
    return retstk[--retidx];
}

// Push a value onto the data stack
void FAST push(int value)
{
    stack[stkidx++] = value;
}

// Pop a value from the data stack
int FAST pop(void)
{
    if (stkidx == 0)
    {
        printf("Stack underflow\n");
        errorcode = RETURN_ERROR;
        return 0;
    }
    return stack[--stkidx];
}

// Pop a double length value from the data stack
int64_t FAST pop2(void)
{
    int64_t value = (int64_t)pop();
    value = (value << 32) | (uint64_t)pop();
    return value;
}

// Push a double length value onto the data stack
void FAST push2(int64_t value)
{
    push((int)(value & 0xffffffff));
    push((int)(value >> 32));
}

// Print a string to the display
void Type(char *str, int len)
{
    while (len-- > 0)
        putchar(*str++);
}

#if 0
// Execute a word's funtion
static inline int Execute(dictentryT *entry)
{
    return (entry->func)(entry);
}
#endif

// Convert an ASCII character to a single digit less than the base
// Return a -l if not a digit or greater than or equal to the base
int getdigit(int value)
{
    if (value >= '0' && value <= '9')
        value -= '0';
    else if (value >= 'a' && value <= 'z')
        value -= 'a' - 10;
    else if (value >= 'A' && value <= 'Z')
        value -= 'A' - 10;
    else
        return -1;
    if (value >= base)
        return -1;
    return value;
}

// Convert a string to a number.  Return 1 if successfully
// converted, or 0 if the string could not be converted.
int getnumber(char *wordptr, int wordlen, int *pval)
{
    int i = 0, value = 0, sflag = 0, digit;
    if (wordptr[0] == '-')
    {
        i = 1;
        sflag = 1;
    }
    else if (wordptr[0] == '+')
        i = 1;
    for (; i < wordlen; i++)
    {
        if ((digit = getdigit(wordptr[i])) < 0) return 0;
        value = (value * base) + digit;
    }
    if (sflag) value = -value;
    *pval = value;
    return 1;
}

// Create a dictionary entry
void Create(char *name, int len, wordfuncP wordfunc)
{
    dictentryT *wordentry;

    here = ALIGN(here);
    wordentry = (dictentryT *)here;
    wordentry->prev = lastword;
    wordentry->does = 0;
    wordentry->func = wordfunc;
    wordentry->flags = 0;
    wordentry->len = len;
    memcpy(wordentry->name, name, len);
    lastword = wordentry;
    here = ALIGN(wordentry->name + len);
}

// Perform a case insensitive comparison of str1 to str2.
// Return 1 if they match, or 0 if they don't match.
int matchstr(char *str1, int len1, char *str2, int len2)
{
    if (len1 != len2) return 0;
    while (len1-- > 0)
    {
        if (toupper((int)(*str1++)) != toupper((int)(*str2++)))
            return 0;
    }
    return 1;
}

// Search the dictionary for a word that matches the string in name.
dictentryT *FindWord(char *name, int len)
{
    dictentryT *wordentry = lastword;

    while (wordentry)
    {
        if (matchstr(wordentry->name, wordentry->len, name, len))
            return wordentry;
        wordentry = wordentry->prev;
    }

    return 0;
}

// Add the entry address of the word in name to the data space.
void AddEntryPtr(char *name)
{
    dictentryT *wordentry = 0;

    if (name) wordentry = FindWord(name, strlen(name));
        
    here = ALIGN(here);
    *((int *)here) = (int)wordentry;
    here += 4;
}

// Find a next character that matches value in the input string.
// Return the value of the input index.
static inline int FindChar(int value)
{
    while (inputidx < inputlen)
    {
        if (inputptr[inputidx] == value) break;
        inputidx++;
    }
    return inputidx;
}

// Skip all characters that match value in the input string.
// Return the value of the input index.
static inline int SkipChar(int value)
{
    while (inputidx < inputlen)
    {
        if (inputptr[inputidx] != value) break;
        inputidx++;
    }
    return inputidx;
}

// Locate the next word in the input string that is delimited by a space.
char *Word(int *num)
{
    int wordindex = SkipChar(' ');
    FindChar(' ');
    *num = inputidx - wordindex;
    if (inputidx < inputlen) inputidx++;
    return inputptr + wordindex;
}

// Set the input string pointer and length to the passed
// parameters and execute the words in the string.
int Evaluate(char *ptr, int len)
{
    int wordlen;
    char *wordptr;
    dictentryT *entry;
    int immediate;
    int postpone = 0;

#ifdef DEBUG
    printf("Evaluate \"");
    Type(ptr, len);
    printf("\"\n");
#endif

    inputptr = ptr;
    inputidx = 0;
    inputlen = len;
    wordptr = Word(&wordlen);
    errorcode = RETURN_NORMAL;
    while (wordlen)
    {
        if (comment)
        {
            if (wordptr[wordlen-1] == ')')
                comment = 0;
            wordptr = Word(&wordlen);
            continue;
        }
        if (wordlen == 1 && *wordptr == '\\') return errorcode;
        entry = FindWord(wordptr, wordlen);
        if (!entry)
        {
            int value;
            if (getnumber(wordptr, wordlen, &value))
            {
                push(value);
                if (compileflag)
                    entry = literalword;
                else
                {
                    wordptr = Word(&wordlen);
                    continue;
                }
            }
            else
            {
                Type(wordptr, wordlen);
                printf(" ?\n");
                if (compileflag)
                {
                    if (lastword) lastword = lastword->prev;
                    compileflag = 0;
                }
                stkidx = 0;
                retidx = 0;
                ctlidx = 0;
                return RETURN_ERROR;
            }
        }

        if (postpone)
        {
            postpone = 0;
            immediate = 0;
        }
        else
            immediate = entry->flags & FLAG_IMMEDIATE;

        postpone = (entry == postponeword);

        if (compileflag && !immediate)
        {
            here = ALIGN(here);
            *((int *)here) = (int)entry;
            here += 4;
            if (entry == literalword)
            {
                *((int *)here) = pop();
                here += 4;
            }
        }
        else
        {
            int retval;
            if ((retval = EXECUTE(entry)) <= RETURN_QUIT)
            {
                if (compileflag)
                {
                    if (lastword) lastword = lastword->prev;
                    compileflag = 0;
                }
                stkidx = 0;
                retidx = 0;
                ctlidx = 0;
                return retval;
            }
        }
        wordptr = Word(&wordlen);
    }
    return errorcode;
}

// Remove carriage-return and line-feed characters from the end of the string.
void RemoveCRLF(char *ptr)
{
    int len = strlen(ptr);
    ptr += len - 1;

    while (len-- > 0)
    {
        if (*ptr != 10 && *ptr != 13) break;
        *ptr-- = 0;
    }
}

// Open the file given by fname, and evaluate all the lines in the file.
// Close the file when done, and return back to the previous input string.
void Include(char *fname)
{
    FILE *infile;
    char inputbuf[100];

    if (!(infile = fopen(fname, "r")))
    {
        printf("Could not open %s\n", fname);
        return;
    }

    instack[inidx++] = inputptr;
    instack[inidx++] = (char *)inputidx;
    instack[inidx++] = (char *)inputlen;

    while (fgets(inputbuf, 100, infile))
    {
        RemoveCRLF(inputbuf);
        if (Evaluate(inputbuf, strlen(inputbuf)) == RETURN_ERROR)
            printf("\n");
    }

    inputlen  = (int)instack[--inidx];
    inputidx  = (int)instack[--inidx];
    inputptr  = instack[--inidx];

    fclose(infile);
}

// Copy the quote-delimited string from the input string to the data area.
void CompileString(void *parm)
{
    char *here0;
    here = ALIGN(here);
    *((int *)here) = (int)parm;
    here += 4;
    here0 = here++;
    while (inputidx < inputlen && inputptr[inputidx] != '"')
        *here++ = inputptr[inputidx++];
    if (inputidx < inputlen) inputidx++;
    *here0 = (char)(here - here0 - 1);
    here = ALIGN(here);
}

#ifdef __PROPELLER__
char getch(void)
{
    int ch;
    FILE save;

    memcpy(&save, stdin, sizeof(FILE));
    setvbuf(stdin, NULL, _IONBF, 0);
    stdin->_flag &= ~_IOCOOKED;

    ch = getchar();

    memcpy(stdin, &save, sizeof(FILE));

    return ch;
}
#else
// Read a character from the input
char getch(void)
{
  struct termios oldt, newt;
  int ch;
  int oldf;

  tcgetattr(STDIN_FILENO, &oldt);
  newt = oldt;
  newt.c_lflag &= ~(ICANON | ECHO);
  tcsetattr(STDIN_FILENO, TCSANOW, &newt);
  oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
  fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);

  ch = getchar();

  tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
  fcntl(STDIN_FILENO, F_SETFL, oldf);

  return ch;
}
#endif

void InitializePfth(void)
{
    here = (char *)codespace;
    lastword = 0;
    stkidx = 0;
    compileflag = 0;
    inidx = 0;
    retidx = 0;
    ctlidx = 0;
    base = 10;
    errorcode = RETURN_NORMAL;
    comment = 0;
}

// Initialize the dictionary and evaluate strings from the input.
#ifndef THIS_IS_MAIN
int pfth(int argc, char **argv)
#else
int main(int argc, char **argv)
#endif
{
    int argi, retval;
    char inputbuf[100];

    printf("pfth 0.22\n\n");
    printf("Initializing\n");
    InitializePfth();    
    InitializeDictionary();
#ifdef __PROPELLER__
    InitializePropWords();
#endif

    for (argi = 1; argi < argc; argi++) Include(argv[argi]);
    
    printf("  ok\n");

    while (gets(inputbuf))
    {
        if ((retval= Evaluate(inputbuf, strlen(inputbuf))) >= RETURN_QUIT)
            printf("  ok\n");
        else if (retval == RETURN_BYE)
        {
            printf("goodbye\n");
            break;
        }
        else
            printf("\n");
    }

    return 0;
}
/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
